# The framework was adapted from https://github.com/theislab/dca/blob/master/dca/api.py

import os, random
import numpy as np
import pandas as pd
import scipy.sparse as sp
import anndata

try:
    import tensorflow as tf
except ImportError:
    raise ImportError('scAAnet requires tensorflow. Please follow instructions'
                      ' at https://www.tensorflow.org/install/ to install it.')

from .train import train
from .network import AE_types

def scAAnet(count,
        ae_type='zinb',
        hidden_size=(128, 10, 128), # network args
        hidden_dropout=0.,
        dispersion='gene-cell',
        activation='relu',
        init='glorot_normal',
        lat_coef=1.,
        network_kwds={},
        epochs=300,               # training args
        reduce_lr=10,
        early_stop=15,
        batch_size=64,
        optimizer='rmsprop',
        learning_rate=0.01,
        random_state=0, 
        verbose=False,
        training_kwds={},
        return_model=False,
        return_loss=False,
        return_info=False,
        warm_up = 20
        ):
    """Single-Cell Archetypal Analysis Neural Network (scAAnet) API.

    Parameters
    ----------
    count : `anndata.AnnData`, `pandas.core.frame.DataFrame` or `numpy.ndarray`
        A dataframe saving raw counts.
    ae_type : `str`, optional. `zipoisson`(default), `zinb`, `nb`, `poisson`, or 'normal'.
        Type of the autoencoder and 'normal' represents an AE with MSE as reconstruction loss.
        Return values and the architecture is determined by the type (e.g. `nb` does not 
        provide dropout probabilities).
    hidden_size : `tuple` or `list`, optional. Default: (128, 10, 128).
        Width of hidden layers.
    hidden_dropout : `float`, `tuple` or `list`, optional. Default: 0.0.
        Probability of weight dropout in the autoencoder (per layer if list
        or tuple).
    dispersion : `str`, optional. `gene` (default) or `gene-cell`.
        If you want the dispersion parameter per gene to be the same across cells,
        leave it as default; otherwise, choosing `gene-cell` will make it differ
        across cells.
    activation : `str`, optional. Default: `relu`.
        Activation function of hidden layers.
    init : `str`, optional. Default: `glorot_uniform`.
        Initialization method used to initialize weights.
    lat_coef : `floar`, optional. Default: 1.0.
        Relative weight of the archetypal loss to the reconstruction loss.
    network_kwds : `dict`, optional.
        Additional keyword arguments for the autoencoder.
    epochs : `int`, optional. Default: 300.
        Number of total epochs in training.
    reduce_lr : `int`, optional. Default: 10.
        Reduces learning rate if loss does not improve in given number of epochs.
    early_stop : `int`, optional. Default: 15.
        Stops training if loss does not improve in given number of epochs.
    batch_size : `int`, optional. Default: 64.
        Number of samples in the batch used for SGD.
    optimizer : `str`, optional. Default: "rmsprop".
        Type of optimization method used for training.
    learning_rate : `float`, optional. Default: 0.01.
        Learning rate to use in the training.
    random_state : `int`, optional. Default: 0.
        Seed for python, numpy and tensorflow.
    verbose : `bool`, optional. Default: `False`.
        If true, prints additional information about training and architecture.
    training_kwds : `dict`, optional.
        Additional keyword arguments for the training process.
    return_model : `bool`, optional. Default: `False`.
        If true, trained autoencoder object is returned. See "Returns".
    return_loss : `bool`, optional. Default: `False`.
        If true, loss information is returned. See "Returns".
    return_info : `bool`, optional. Default: `False`.
        If true, all additional parameters of scAAnet are returned such as dropout
        probabilities and estimated dispersion values, in case that autoencoder 
        is of type zinb or zinb-conddisp.
    warm_up : `int`. Default: 20.
        Number of epochs for warm-up. During warm-ups, the weights of Z_fixed layer
        is not trained.

    Returns
    -------
    If `return_model` is true, reconstruction, usage, archetype matrices and a trained model 
    are returned. Otherwise, only the three matrices are returned as a tuple.
    If `return_loss` is true, two loss values are returned, including reconstruction loss (neg
    log-likelihood) and reconstruction + archetypal loss.
    """

    # set seed for reproducibility
    random.seed(random_state)
    np.random.seed(random_state)
    tf.random.set_seed(random_state)
    os.environ['PYTHONHASHSEED'] = '0'
    
    if isinstance(count, anndata.AnnData):
        if sp.issparse(count.X):
            count = count.X.todense()
        else:
            count = count.X
    count = np.asmatrix(count).astype('float32')

    TPM = count/count.sum(axis=1)
    lib_size = count.sum(axis=1)
    
    network_kwds = {**network_kwds,
        'hidden_size': hidden_size,
        'hidden_dropout': hidden_dropout,
        'dispersion': dispersion,
        'activation': activation,
        'lat_coef': lat_coef,
        'init': init
    }

    input_size = output_size = count.shape[1]
    net = AE_types[ae_type](input_size=input_size,
                            output_size=output_size,
                            **network_kwds)

    net.save()
    net.build_enc()
    net.build_dec()

    training_kwds = {**training_kwds,
        'epochs': epochs,
        'reduce_lr': reduce_lr,
        'early_stop': early_stop,
        'batch_size': batch_size,
        'optimizer': optimizer,
        'verbose': verbose,
        'learning_rate': learning_rate,
        'ae_type': ae_type,
        'warm_up': warm_up,
        'Z_fixed_idx': len(hidden_size) + 1
    }

    train(count, TPM, lib_size, net, **training_kwds)
    preds = net.predict(TPM, lib_size, return_info)
    
    outputs = net.model([TPM, lib_size], training=False)
    loss_value = net.loss(count, outputs)
    loss_value_all = loss_value + sum(net.model.losses)
    losses = {"loss": loss_value.numpy(), "loss_all": loss_value_all.numpy()}
    if return_model:
        if return_loss:
            return preds, losses, net
        else:
            return preds, net
    else:
        if return_loss:
            return preds, losses
        else:
            return preds



import os
import numpy as np
import tensorflow as tf
import tensorflow.keras.optimizers as opt

def train(count, nor_count, lib_size, network, output_dir=None, optimizer='rmsprop', learning_rate=None,
          epochs=300, reduce_lr=10, output_subset=None, use_raw_as_output=True,
          early_stop=15, batch_size=256, clip_grad=5., save_weights=False,
          validation_split=0.1, tensorboard=False, verbose=True, ae_type='normal',
          warm_up = 0, Z_fixed_idx = None,
          **kwds):

    model = network.model
    loss = network.loss
    if output_dir is not None:
        os.makedirs(output_dir, exist_ok=True)

    if learning_rate is None:
        optimizer = opt.RMSprop(clipvalue=clip_grad)
    else:
        optimizer = opt.RMSprop(lr=learning_rate, clipvalue=clip_grad)


    train_dataset = tf.data.Dataset.from_tensor_slices((count, nor_count, lib_size))
    train_dataset = train_dataset.shuffle(buffer_size=count.shape[0],
                                          reshuffle_each_iteration=True).batch(batch_size)

    @tf.function
    def train_step_warm_up(count, nor_count, lib_size):
        # Open a GradientTape to record the operations run
        # during the forward pass, which enables auto-differentiation.
        
        with tf.GradientTape() as tape:
            # Run the forward pass of the layer.
            # The operations that the layer applies to its inputs are going to be recorded on the GradientTape.
            outputs = model([nor_count, lib_size], training=True)
            # Compute the loss value for this minibatch.
            if ae_type == 'normal':
                loss_value = loss(nor_count, outputs)
            else:
                loss_value = loss(count, outputs)
            # Add archetypal loss term
            loss_value += sum(model.losses)
        # Use the gradient tape to automatically retrieve
        # the gradients of the trainable variables with respect to the loss.
        var_list = model.trainable_weights[:Z_fixed_idx] + model.trainable_weights[(Z_fixed_idx+1):]
        grads = tape.gradient(loss_value, var_list)
        # Run one step of gradient descent by updating
        # the value of the variables to minimize the loss.
        optimizer.apply_gradients(zip(grads, var_list))
        
        return loss_value
    
    @tf.function
    def train_step(count, nor_count, lib_size):
        # Open a GradientTape to record the operations run
        # during the forward pass, which enables auto-differentiation.
        
        with tf.GradientTape() as tape:
            # Run the forward pass of the layer.
            # The operations that the layer applies to its inputs are going to be recorded on the GradientTape.
            outputs = model([nor_count, lib_size], training=True)
            # Compute the loss value for this minibatch.
            if ae_type == 'normal':
                loss_value = loss(nor_count, outputs)
            else:
                loss_value = loss(count, outputs)
            # Add archetypal loss term
            loss_value += sum(model.losses)
        # Use the gradient tape to automatically retrieve
        # the gradients of the trainable variables with respect to the loss.
        grads = tape.gradient(loss_value, model.trainable_weights)
        optimizer.apply_gradients(zip(grads, model.trainable_weights))
 
        return loss_value

    best_weights = None
    wait_early_stop = wait_reduce_lr = 0
    stopped_epoch = 0
    best = np.Inf
    factor = 0.5
    min_lr = 0.001

    for epoch in range(epochs):

        # Iterate over the batches of the dataset.
        for step, (count, nor_count, lib_size) in enumerate(train_dataset):
            #print('Step: ', step)
            if epoch < warm_up:
                loss_value = train_step_warm_up(count, nor_count, lib_size)
            else:
                loss_value = train_step(count, nor_count, lib_size)
        
        current = loss_value
        
        if epoch % 50 == 0:
            print("Training loss at epoch %d (for the last batch): %.4f"
                  % (epoch, float(loss_value)))
            print("Learning rate is: %.6f" % float(tf.keras.backend.get_value(optimizer.lr)))
            print(model.trainable_weights[Z_fixed_idx])
                
        if np.less(current, best):
            best = current
            wait_early_stop = wait_reduce_lr = 0
            best_weights = model.get_weights()
        else:
            wait_early_stop += 1
            wait_reduce_lr += 1
            if wait_early_stop >= early_stop:
                stopped_epoch = epoch
                print("Restoring model weights from the end of the best epoch.")
                model.set_weights(best_weights)
            if wait_reduce_lr >= reduce_lr:
                wait_reduce_lr = 0
                lr = float(tf.keras.backend.get_value(optimizer.lr))
                if lr*factor > min_lr:
                    tf.keras.backend.set_value(optimizer.lr, lr*factor)
                else:
                    tf.keras.backend.set_value(optimizer.lr, min_lr)
                
        if stopped_epoch > 0:
            print("Epoch %05d: early stopping" % (stopped_epoch + 1))
            break
            
    return None



import setuptools

with open("./README.md", "r") as fh:
  long_description = fh.read()

setuptools.setup(
  name="scAAnet",
  version="1.0.0",
  author="Yuge Wang",
  author_email="wangyuge22@qq.com",
  description="An implementation of nonlinear archetypal analysis on single-cell RNA-seq data through autoencoder",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/AprilYuge/scAAnet_latest",
  install_requires = [
    'numpy',
    'pandas',
    'scipy',
    'tensorflow',
    'sklearn',
    'fastcluster',
    'statsmodels',
    'seaborn',
    'matplotlib',
    'anndata'
  ],
  packages=['scAAnet'],
  classifiers=[
  "Programming Language :: Python :: 3",
  "License :: OSI Approved :: MIT License",
  "Operating System :: OS Independent",
  ],
)


import numpy as np
import pandas as pd
from statsmodels.stats.multitest import multipletests
import statsmodels.formula.api as smf 
import statsmodels.api as sm

import matplotlib.pyplot as plt
import seaborn as sns
import matplotlib.gridspec as gridspec

import scipy
from scipy.cluster.hierarchy import cophenet, leaves_list
from scipy.spatial.distance import squareform, pdist
import scipy.stats as stats
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_score
import fastcluster as fc
import scipy.sparse as sp
import anndata

def save_df_to_npz(obj, filename):
    np.savez_compressed(filename, data=obj.values, index=obj.index.values, columns=obj.columns.values)

def load_df_from_npz(filename):
    with np.load(filename, allow_pickle = True) as f:
        obj = pd.DataFrame(**f)
    return obj

# Program specific DEG
def programDEG(count, usage, test_use = 'nb', offset = True, p_cor = 'bonferroni', lib_size=None, min_counts_per_cell=None):
    """

    Parameters
    ----------
    count : `pandas.core.frame.DataFrame` or `anndata.AnnData`
        A dataframe saving raw counts with rows being cells and columns being genes.
        Or an anndata object with raw count data in .X
    usage: `numpy.ndarray` or `pd.core.frame.DataFrame`
        A numpy array or a pandas dataframe saving usage and each column corresponds to a program.
    test.use: `str`, optional. Default: `nb`.
        This is the type of GLM regression used for DEG, other choices are `nb_naive` and `poisson`.
    p_cor: `str`, optional. Default: `bonferroni`.
        Method used for p value adjustment. Another choice is `benjamini-hochberg`.

    Returns
    -------
    A dictionary with each element corresponds to results for a program. Columns of each 
    dataframe are coefficient estimates, p values, and adjusted p values. 
    """
    
    results = {}
    
    if lib_size is None:
        lib_size = count.sum(1)
        
    if isinstance(count, anndata.AnnData):
        cells = count.obs.index
        genes = count.var.index
        if sp.issparse(count.X):   
            count = count.X.todense()
        else:
            count = count.X
        count = pd.DataFrame(count)
        count.index = cells
        count.columns = genes
    
    if isinstance(usage, pd.core.frame.DataFrame):
        usage = usage.values
    
    # Filter out cells with too small library size
    if min_counts_per_cell is None:
        min_counts_per_cell=1/20 * count.shape[1]
    count = count[lib_size >= min_counts_per_cell]
    usage = usage[lib_size >= min_counts_per_cell,:]
    lib_size = lib_size[lib_size >= min_counts_per_cell]
    
    for i in range(usage.shape[1]):
        
        # Save results for current program in result
        result = []
        count_a = 0

        for j in range(count.shape[1]):
            
            # Check if the current gene has 0 variation across cells
            if np.sum(count.iloc[:,j]>0)<=count.shape[0]/500:
                continue
            
            # Negatve binomial regression for the ith program and jth gene
            data = pd.DataFrame({'usage': usage[:,i], 'gene': count.iloc[:,j], 'log_lib_size': np.log(lib_size)})
            
            if offset:
                offs = data['log_lib_size']
                formula = "gene ~ usage"
            else:
                offs = None
                formula = "gene ~ usage + log_lib_size"
            
            if test_use == 'poisson':
                model = smf.glm(formula=formula, data=data, offset=offs, family=sm.families.Poisson()).fit()
            elif test_use == 'nb':      
                # EStimate alpha
                model_p = smf.glm(formula=formula, data=data, offset=offs, family=sm.families.Poisson()).fit()
                data['bb_lambda'] = model_p.mu
                data['aux_ols_dep'] = data.apply(lambda x: ((x['gene'] - x['bb_lambda'])**2 - x['bb_lambda']) / x['bb_lambda'], axis=1)
                model_aux_ols = smf.ols(formula="aux_ols_dep ~ bb_lambda - 1", data=data).fit()
            
                # print alpha
                #print(model_aux_ols.params[0])
                if model_aux_ols.params[0] < 0:
                    count_a += 1
                    
                # Truncate alpha
                alpha = 2 if model_aux_ols.params[0] > 2 else model_aux_ols.params[0]
                alpha = 0.01 if alpha < 0.01 else alpha
            
                # Fit NB regression model
                model = smf.glm(formula=formula, data=data, offset=offs, 
                                family=sm.families.NegativeBinomial(alpha=alpha)).fit()
            elif test_use == 'nb_naive':
                # Fit NB regression model
                model = smf.glm(formula=formula, data=data, offset=offs, 
                                family=sm.families.NegativeBinomial()).fit()

            # Append results
            result.append([count.columns[j], model.params[1], model.pvalues[1], model.tvalues[1]])
            
        result = pd.DataFrame(result, columns=['gene', 'coef.est', 'p.val', 'z.score'])
        result.loc[np.isnan(result['p.val']),'p.val'] = 1
        
        # Calculate adjusted p vals
        if p_cor == 'benjamini-hochberg':
            _, p_vals_adj, _, _ = multipletests(result['p.val'], alpha=0.05, method='fdr_bh')
            result['p.val.adj'] = p_vals_adj
        elif p_cor == 'bonferroni':
            result['p.val.adj'] = np.minimum(result['p.val']*result.shape[0], 1.0)
     
        # Add result
        results['gep_%s'%(i+1)] = result    
        
        if test_use == 'nb':
            print('Number of negative alpha is %s' %count_a)
    
    return results

# Ranked gene scores in each GEP
def plotRankedGeneScore(spectra, ncol=4):
    nrow = int(np.ceil(spectra.shape[1]/ncol))
    fig = plt.figure(figsize=(ncol*2, nrow*2), dpi=800)
    gs = gridspec.GridSpec(nrow, ncol, fig, 0, 0, 1, 1, hspace=.8)
    for (i,program) in enumerate(spectra.columns):
        ax = fig.add_subplot(gs[int(i/ncol), i%ncol], frameon=False, rasterized=False)
        x_ind = np.arange(1, spectra.shape[0]+1)
        spectra_sub = spectra.sort_values(by=spectra.columns[i], axis=0, ascending=False).iloc[:,i]
        value = spectra_sub.to_numpy()
        ax.scatter(x_ind, value, label=None, s=5, alpha=.8)
        #ax.set_xlabel('Ranked gene list')
        #ax.set_ylabel('Relative expression')
        ax.set_title('GEP %s'%program)
        
# Density plot of gene scores in each GEP
def plotDensity(spectra, ncol=4):
    nrow = int(np.ceil(spectra.shape[1]/ncol))
    fig = plt.figure(figsize=(ncol*2, nrow*2), dpi=800)
    gs = gridspec.GridSpec(nrow, ncol, fig, 0, 0, 1, 1, hspace=.8)
    for (i,program) in enumerate(spectra.columns):
        ax = fig.add_subplot(gs[int(i/ncol), i%ncol], frameon=False, rasterized=False)
        spectra_sub = spectra.iloc[:,i]
        value = spectra_sub.to_numpy()
        sns.distplot(value, hist=True, bins=30, ax=ax)
        ax.set_title('GEP %s'%program)
        
# Density plot of gene scores fitted by Gamma distributions in each GEP
def plotDensityGamma(spectra, ncol=4, thrP=0.975):
    selected_genes = {}
    nrow = int(np.ceil(spectra.shape[1]/ncol))
    fig = plt.figure(figsize=(ncol*2, nrow*2), dpi=800)
    gs = gridspec.GridSpec(nrow, ncol, fig, 0, 0, 1, 1, hspace=.8)
    for (i,program) in enumerate(spectra.columns):
        ax = fig.add_subplot(gs[int(i/ncol), i%ncol], frameon=False, rasterized=False)
        spectra_sub = spectra.iloc[:,i]
        value = spectra_sub.to_numpy()
        value[value==0] += 1e-8
        x = np.linspace(1e-3, value.max(), 100)
        # Fit a Gamma distribution
        gamma = stats.gamma
        param = gamma.fit(value, floc=0)
        pdf_fitted = gamma.pdf(x, *param)
        ax.plot(x, pdf_fitted, color='r')
        # Get gene score cutoff
        cutoff = gamma.ppf(thrP, *param)
        ax.axvline(cutoff, color='grey', linestyle='--')
        # Histogram
        ax.hist(value, bins=30, density=True)
        ax.set_title('GEP %s'%program)
        # Number of genes pass the cutoff
        num = np.sum(value > cutoff)
        ax.set_xlabel('%d genes with score > %.3f' %(num, cutoff))
        # Add genes to dictionary
        spectra_sub = spectra.sort_values(by=spectra.columns[i], axis=0, ascending=False).iloc[:,i]
        selected_genes['GEP %d'%(i+1)] = spectra_sub.index[:num]
    
    return selected_genes

# Calculate gene scores based on the equation given in cisTopic
def geneScore(spectra, scale=True):
    gene_score = spectra*(np.log(spectra+1e-5).sub(np.mean(np.log(spectra+1e-5), axis=1), axis='index'))
    if scale:
        gene_score = gene_score.sub(gene_score.min(axis=0), axis='columns').div(gene_score.max(axis=0)-gene_score.min(axis=0), 
                                                                                axis='columns')
    return gene_score

# Assess the stability of inferred archetypes through differnt Ks
def StabilityArchetype(k_min, k_max, step, n_rep, dists, filepath, savepath=None, l2 = False):
    
    '''
    Borrowed from Alexandrov et al. 2013: Deciphering signatures of mutational processes 
    operative in human cancer and Kotliar et al. 2019: Identifying gene expression programs
    of cell-type identify and cellular activity with single-cell RNA-seq
    '''
    
    K = np.arange(k_min, k_max+1, step)
    results = []
    
    for k in K:
        for dist in dists:
            # Read all spectra results
            arch = load_df_from_npz(filepath%(dist, 1, k))
            arch = arch.transpose()
            for rs in np.arange(2, n_rep+1):
                arch_new = load_df_from_npz(filepath%(dist, rs, k))
                arch = pd.concat([arch, arch_new.transpose()])
    
            if l2:
                arch = (arch.T/np.sqrt((arch**2).sum(axis=1))).T
                
            if savepath:
                save_df_to_npz(arch, savepath%(dist, k))
        
            # K-means clustering on all spectra
            kmeans_model = KMeans(n_clusters=k, n_init=10, random_state=1)
            kmeans_model.fit(arch)
            kmeans_cluster_labels = pd.Series(kmeans_model.labels_+1, index=arch.index)

            # Compute the silhouette score
            stability = silhouette_score(arch.values, kmeans_cluster_labels, metric='euclidean')
            results.append([k, dist, stability])
    
    results = pd.DataFrame(results, columns=["K", "Method", "Stability"])
    
    return results

# Assess the stability of inferred archetypes through differnt Ks
def StabilityUsage(k_min, k_max, step, n_rep, dists, n_cells, filepath_usage, filepath_consensus=None, 
                   n_cells_max=20000, cluster=True):
    
    '''
    Borrowed from Brunet et al. 2004: Metagenes and molecular pattern discovery using
    matrix factorization
    '''
    np.random.seed(520)
    
    K = np.arange(k_min, k_max+1, step)
    results = []
    
    if n_cells > n_cells_max:
        cells_choice = np.random.choice(n_cells, n_cells_max, replace=False)
        n_cells = n_cells_max
    else:
        cells_choice = np.arange(n_cells)
    
    for k in K:
        for dist in dists:
            d = np.zeros(int(scipy.special.comb(n_cells, 2)))
            # Read all cell usage results
            for rs in np.arange(1, n_rep+1):
                usage = load_df_from_npz(filepath_usage%(dist, rs, k))
                usage = usage.to_numpy()
                usage = usage[cells_choice,]               
                if cluster:
                    assign = usage.argmax(1)
                    usage = np.zeros_like(usage)
                    usage[np.arange(len(usage)), assign] = 1
                    d += scipy.spatial.distance.pdist(usage, 'braycurtis')
                else:
                    d += scipy.spatial.distance.pdist(usage)
            
            d = d/n_rep
            
            if filepath_consensus:
                if cluster:
                    # Save consensus clustering matrix
                    np.save(filepath_consensus%(dist, k), 1-d)
                else:
                    # Save distance matrix
                    np.save(filepath_consensus%(dist, k), d)
            
            # Hierarchical clustering using distance d
            HC = fc.linkage(d, method='average')
            cophen_d = cophenet(HC)
            
            # Compute Cophenetic correlation coefficient
            cophen_corr = np.corrcoef(d, cophen_d)[0,1]
            results.append([k, dist, cophen_corr])
    
    results = pd.DataFrame(results, columns=["K", "Method", "Stability"])
    
    return results

# Plot consensus clustering matrix of inferred usage
def PlotConsensusMat(filepath, savepath=None, colormap='rocket', n_cells_max=6000, dpi=600):
    np.random.seed(520)
    cvec = np.load(filepath)    
    cmat = squareform(cvec)
    if len(cmat) > n_cells_max:
        cells_choice = np.random.choice(len(cmat), n_cells_max, replace=False)
        cmat = cmat[cells_choice,][:, cells_choice]
        cvec = scipy.spatial.distance.squareform(cmat)
    HC = fc.linkage(1-cvec, method='average')
    cg = sns.clustermap(cmat, row_linkage=HC, col_linkage=HC, cbar_pos=(1, .2, .03, .4), 
                        yticklabels=False, xticklabels=False, cmap=colormap)
    cg.ax_row_dendrogram.set_visible(False)
    cg.ax_col_dendrogram.set_visible(False)
    if savepath:
        plt.savefig(savepath, dpi=dpi)
        plt.close()

# Plot distance clustering matrix of inferred GEPs
def PlotArchDistMat(filepath, savepath=None, colormap='rocket', dpi=600):
    arch = load_df_from_npz(filepath)
    
    # K-means clustering on all spectra
    kmeans_model = KMeans(n_clusters=k, n_init=10, random_state=1)
    kmeans_model.fit(arch)
    kmeans_cluster_labels = pd.Series(kmeans_model.labels_+1, index=arch.index)
    
    topics_dist = squareform(pdist(arch.values))
    
    spectra_order = []
    for cl in sorted(set(kmeans_cluster_labels)):
        cl_filter = kmeans_cluster_labels==cl

        if cl_filter.sum() > 1:
            cl_dist = squareform(topics_dist[cl_filter, :][:, cl_filter])
            cl_dist[cl_dist < 0] = 0 #Rarely get floating point arithmetic issues
            cl_link = fc.linkage(cl_dist, 'average')
            cl_leaves_order = leaves_list(cl_link)

            spectra_order += list(np.where(cl_filter)[0][cl_leaves_order])
        else:
            ## Corner case where a component only has one element
            spectra_order += list(np.where(cl_filter)[0])
            
    width_ratios = [0.5, 9, 0.5, 0.5]
    height_ratios = [0.5, 9]
    fig = plt.figure(figsize=(sum(width_ratios), sum(height_ratios)))
    gs = gridspec.GridSpec(len(height_ratios), len(width_ratios), fig,
                            0.01, 0.01, 0.98, 0.98,
                            height_ratios=height_ratios,
                            width_ratios=width_ratios,
                            wspace=0, hspace=0)

    dist_ax = fig.add_subplot(gs[1,1], xticks=[], yticks=[], frameon=False)

    D = topics_dist[spectra_order, :][:, spectra_order]
    dist_im = dist_ax.imshow(D, interpolation='none', cmap=colormap, aspect='auto', rasterized=True)

    left_ax = fig.add_subplot(gs[1,0], xticks=[], yticks=[], frameon=False)
    left_ax.imshow(kmeans_cluster_labels.values[spectra_order].reshape(-1, 1),
                    interpolation='none', cmap='Spectral', aspect='auto', rasterized=True)

    top_ax = fig.add_subplot(gs[0,1], xticks=[], yticks=[], frameon=False)
    top_ax.imshow(kmeans_cluster_labels.values[spectra_order].reshape(1, -1),
                  interpolation='none', cmap='Spectral', aspect='auto', rasterized=True)

    cbar_gs = gridspec.GridSpecFromSubplotSpec(3, 1, subplot_spec=gs[1, 3],
                                               height_ratios=[0.5,1,0.5], wspace=0, hspace=0)
    cbar_ax = fig.add_subplot(cbar_gs[1,0], frameon=False, title='')
    fig.colorbar(dist_im, cax=cbar_ax, ticks=np.linspace(D.min(), D.max(), 6).round(2), 
                 orientation='vertical').outline.set_visible(False)
    
    if savepath:
        plt.savefig(savepath, dpi=dpi, bbox_inches = 'tight')
        plt.close()



# This script was adapted from https://github.com/theislab/dca/blob/master/dca/loss.py

import numpy as np
import tensorflow as tf

def _nan2zero(x):
    return tf.where(tf.math.is_nan(x), tf.zeros_like(x), x)

def _nan2inf(x):
    return tf.where(tf.math.is_nan(x), tf.zeros_like(x)+np.inf, x)

def _nelem(x):
    nelem = tf.reduce_sum(tf.cast(~tf.math.is_nan(x), tf.float32))
    return tf.cast(tf.where(tf.math.equal(nelem, 0.), 1., nelem), x.dtype)

def _reduce_mean(x):
    nelem = _nelem(x)
    x = _nan2zero(x)
    return tf.math.divide(tf.math.reduce_sum(x), nelem)

def poisson_loss(y_true, y_pred, mean=True, masking=False):
    
    y_pred = tf.cast(y_pred, tf.float32)
    y_true = tf.cast(y_true, tf.float32)

    y_true = _nan2zero(y_true)

    ret = y_pred - y_true*tf.math.log(y_pred+1e-10) + tf.math.lgamma(y_true+1.0)
    
    if mean:
        if masking:
            ret = _reduce_mean(ret)
        else:
            ret = tf.math.reduce_mean(ret)

    return ret

def ZIPoisson_loss(y_true, outputs, ridge_lambda=0.0,
              mean=True, masking=False, scale_factor=1.0, debug=False):
    
    y_pred, pi = outputs
    softplus_pi = tf.math.softplus(-pi)  #  uses log(sigmoid(x)) = -softplus(-x)

    # reuse existing Poisson neg.log.lik.
    # mean is always False here, because everything is calculated
    # element-wise. we take the mean only in the end
    poisson_case = poisson_loss(y_true, y_pred, mean=False) - (-softplus_pi - pi)

    y_true = tf.cast(y_true, tf.float32)
    y_pred = tf.cast(y_pred, tf.float32) * scale_factor

    zero_case = -tf.math.softplus(-pi - y_pred) + softplus_pi
    result = tf.where(tf.math.less(y_true, 1e-8), zero_case, poisson_case)
    ridge = ridge_lambda*tf.math.square(pi)
    result += ridge

    if mean:
        if masking:
            result = _reduce_mean(result)
        else:
            result = tf.math.reduce_mean(result)

    result = _nan2inf(result)

    if debug:
        tf.summary.histogram('poisson_case', poisson_case)
        tf.summary.histogram('zero_case', zero_case)
        tf.summary.histogram('ridge', ridge)

    return result

def NB_loss(y_true, outputs, mean=True,
            masking=False, scale_factor=1.0, debug=False):
    eps = 1e-10
    y_pred, theta = outputs

    y_true = tf.cast(y_true, tf.float32)
    y_pred = tf.cast(y_pred, tf.float32) * scale_factor

    if masking:
        nelem = _nelem(y_true)
        y_true = _nan2zero(y_true)

    # Clip theta
    theta = tf.math.minimum(theta, 1e6)

    t1 = tf.math.lgamma(theta+eps) + tf.math.lgamma(y_true+1.0) - tf.math.lgamma(y_true+theta+eps)
    t2 = (theta+y_true) * tf.math.log(1.0 + (y_pred/(theta+eps))) + (y_true * (tf.math.log(theta+eps) - tf.math.log(y_pred+eps)))

    if debug:
        assert_ops = [
            tf.verify_tensor_all_finite(y_pred, 'y_pred has inf/nans'),
            tf.verify_tensor_all_finite(t1, 't1 has inf/nans'),
            tf.verify_tensor_all_finite(t2, 't2 has inf/nans')]

        tf.summary.histogram('t1', t1)
        tf.summary.histogram('t2', t2)

        with tf.control_dependencies(assert_ops):
            final = t1 + t2

    else:
        final = t1 + t2

    final = _nan2inf(final)

    if mean:
        if masking:
            final = tf.math.divide(tf.math.reduce_sum(final), nelem)
        else:
            final = tf.math.reduce_mean(final)


    return final

def ZINB_loss(y_true, outputs, ridge_lambda=0.0,
              mean=True, masking=False, scale_factor=1.0, debug=False):
    eps = 1e-10
    y_pred, theta, pi = outputs
    softplus_pi = tf.math.softplus(-pi)  #  uses log(sigmoid(x)) = -softplus(-x)

    # reuse existing NB neg.log.lik.
    # mean is always False here, because everything is calculated
    # element-wise. we take the mean only in the end
    nb_case = NB_loss(y_true, [y_pred, theta], mean=False) - (-softplus_pi - pi)

    y_true = tf.cast(y_true, tf.float32)
    y_pred = tf.cast(y_pred, tf.float32) * scale_factor
    theta = tf.math.minimum(theta, 1e6)

    pi_theta_log = -pi + theta * (tf.math.log(theta+eps) - tf.math.log(theta+y_pred+eps))
    zero_case = -tf.math.softplus(pi_theta_log) + softplus_pi
    
    result = tf.where(tf.math.less(y_true, 1e-8), zero_case, nb_case)
    ridge = ridge_lambda*tf.math.square(pi)
    result += ridge

    if mean:
        if masking:
            result = _reduce_mean(result)
        else:
            result = tf.math.reduce_mean(result)

    result = _nan2inf(result)

    if debug:
        tf.summary.histogram('nb_case', nb_case)
        tf.summary.histogram('zero_case', zero_case)
        tf.summary.histogram('ridge', ridge)

    return result


# The framework was adapted from https://github.com/theislab/dca/blob/master/dca/network.py

import numpy as np
import os
import pickle

import tensorflow as tf
from tensorflow.keras.layers import Input, Dense, Dropout, Lambda, Concatenate
from tensorflow.keras import Model
from tensorflow.keras.regularizers import l1_l2

from .loss import poisson_loss, NB_loss, ZINB_loss, ZIPoisson_loss

ColwiseMultLayer = Lambda(lambda l: l[0]*tf.reshape(l[1], (-1,1)), name="mean")
DispAct = lambda x: tf.clip_by_value(tf.nn.softplus(x), 1e-4, 1e4)

class DispLayer(tf.keras.layers.Layer):
    def __init__(self, units=2000):
        super(DispLayer, self).__init__()
        self.w = tf.Variable(initial_value=tf.random.normal([units]), trainable=True, name='dispersion')

    def call(self, inputs):
        return DispAct(self.w)

class ZFixedLayer(tf.keras.layers.Layer):
    def __init__(self, dim_latent_space):
        super(ZFixedLayer, self).__init__(name='z_fixed')
        self.w = tf.Variable(initial_value=tf.convert_to_tensor(create_z_fixed(dim_latent_space), tf.keras.backend.floatx()), 
                             trainable=True, name='z_fixed')

    def call(self, inputs):
        return tf.matmul(inputs, self.w)

# We borrowed the following function from https://github.com/bmda-unibas/DeepArchetypeAnalysis/blob/master/AT_lib/lib_at.py
def create_z_fixed(dim_latent_space):
    """
    Creates Coordinates of the Simplex spanned by the Archetypes.
    The simplex will have its centroid at 0.
    The sum of the vertices will be zero.
    The distance of each vertex from the origin will be 1.
    The length of each edge will be constant.
    The dot product of the vectors defining any two vertices will be - 1 / M.
    This also means the angle subtended by the vectors from the origin
    to any two distinct vertices will be arccos ( - 1 / M ).
    :param dim_latent_space:
    :return:
    """

    z_fixed_t = np.zeros([dim_latent_space, dim_latent_space + 1])

    for k in range(0, dim_latent_space):
        s = 0.0
        for i in range(0, k):
            s = s + z_fixed_t[i, k] ** 2

        z_fixed_t[k, k] = np.sqrt(1.0 - s)

        for j in range(k + 1, dim_latent_space + 1):
            s = 0.0
            for i in range(0, k):
                s = s + z_fixed_t[i, k] * z_fixed_t[i, j]

            z_fixed_t[k, j] = (-1.0 / float(dim_latent_space) - s) / z_fixed_t[k, k]
            z_fixed = np.transpose(z_fixed_t)
    return z_fixed

class Autoencoder():   
    def __init__(self,
                 input_size,
                 output_size=None,
                 hidden_size=(64, 32, 64),
                 dispersion = 'gene',
                 lat_coef=1.,
                 l2_coef=0.,
                 l1_coef=0.,
                 l2_enc_coef=0.,
                 l1_enc_coef=0.,
                 ridge=0.,
                 hidden_dropout=0.,
                 input_dropout=0.,
                 activation='relu',
                 init='glorot_normal',
                 file_path=None,
                 debug=False):

        self.input_size = input_size
        self.output_size = output_size
        self.hidden_size = hidden_size
        self.dispersion = dispersion
        self.lat_coef = lat_coef
        self.l2_coef = l2_coef
        self.l1_coef = l1_coef
        self.l2_enc_coef = l2_enc_coef
        self.l1_enc_coef = l1_enc_coef
        self.ridge = ridge
        self.hidden_dropout = hidden_dropout
        self.input_dropout = input_dropout
        self.activation = activation
        self.init = init
        self.loss = None
        self.file_path = file_path
        self.extra_models = {}
        self.model = None
        self.encoder = None
        self.decoder = None
        self.input_layer = None
        self.sf_layer = None
        self.debug = debug

        if self.output_size is None:
            self.output_size = input_size
            
        if isinstance(self.hidden_dropout, list):
            assert len(self.hidden_dropout) == len(self.hidden_size)
        else:
            self.hidden_dropout = [self.hidden_dropout]*len(self.hidden_size)

    def build_enc(self):

        self.input_layer = Input(shape=(self.input_size,), name='nor_count')
        self.sf_layer = Input(shape=(1,), name='lib_size')
        last_hidden = self.input_layer
        
        if self.input_dropout > 0.0:
            last_hidden = Dropout(self.input_dropout, name='input_dropout')(last_hidden)
            
        for i, (hid_size, hid_drop) in enumerate(zip(self.hidden_size, self.hidden_dropout)):
            center_idx = int(np.floor(len(self.hidden_size) / 2.0))
            self.center_idx = center_idx
            if i == center_idx:
                layer_name = 'center'
                stage = 'center'  # let downstream know where we are
                self.num_at = hid_size
                self.dim_latent_space = self.num_at - 1
            elif i < center_idx:
                layer_name = 'enc%s' % i
                stage = 'encoder'
            else:
                #layer_name = 'dec%s' % (i-center_idx)
                stage = 'decoder'
                break

            # use encoder-specific l1/l2 reg coefs if given
            if self.l1_enc_coef != 0. and stage in ('center', 'encoder'):
                l1 = self.l1_enc_coef
            else:
                l1 = self.l1_coef

            if self.l2_enc_coef != 0. and stage in ('center', 'encoder'):
                l2 = self.l2_enc_coef
            else:
                l2 = self.l2_coef
                
            if stage == 'center': # yuge
                fc_a = Dense(hid_size, activation='softmax', kernel_initializer=self.init,
                                kernel_regularizer=l1_l2(l1, l2),
                                name=layer_name)(last_hidden)
                fc_b_t = Dense(hid_size, activation=None, kernel_initializer=self.init,
                                kernel_regularizer=l1_l2(l1, l2),
                                name="%s_b_t"%layer_name)(last_hidden)
                # Add archetype regularization loss
                #z_fixed = tf.eye(self.num_at)
                #self.z_fixed = create_z_fixed(self.dim_latent_space)
                #mu_t = tf.matmul(fc_a, self.z_fixed)
                
                mu_t = ZFixedLayer(self.dim_latent_space)(fc_a)
                fc_b = tf.nn.softmax(tf.transpose(fc_b_t), 1)
                self.z_predicted = tf.matmul(fc_b, mu_t)
                
                #self.arch_loss = tf.math.reduce_mean(tf.math.square(self.z_fixed - z_predicted))
                #last_hidden = fc_a
                last_hidden = mu_t
            else:
                last_hidden = Dense(hid_size, activation = tf.nn.leaky_relu,
                                    kernel_initializer=self.init,
                                    kernel_regularizer=l1_l2(l1, l2),
                                    name=layer_name)(last_hidden)
                if hid_drop > 0.0:
                    last_hidden = Dropout(hid_drop, name='%s_drop'%layer_name)(last_hidden)
        
        self.encoder_output = last_hidden
        
    def build_dec(self):
        
        last_hidden = Dense(self.hidden_size[(self.center_idx+1)], 
                            kernel_initializer=self.init, 
                            activation=tf.nn.leaky_relu,
                            kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                            name='dec0')(self.encoder_output)
        
        if len(self.hidden_size) > (self.center_idx+2):
            for i, hid_size in enumerate(self.hidden_size[(self.center_idx+2):]):
                last_hidden = Dense(hid_size, activation=tf.nn.leaky_relu,
                                    kernel_initializer=self.init,
                                    kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                                    name="dec%s"%(i+1))(last_hidden)
        
        self.decoder_output = last_hidden
        self.dec_layer_num = len(self.hidden_size) - self.center_idx - 1
        self.build_output()

    def build_output(self):

        self.loss = tf.keras.losses.MeanSquaredError()
        mean = Dense(self.output_size, kernel_initializer=self.init, activation="softmax",
                     kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                     name='scaled_mean')(self.decoder_output)
        #output = ColwiseMultLayer([mean, self.ls_layer])

        # keep unscaled output as an extra model
        self.extra_models['mean_norm'] = Model(inputs=self.input_layer, outputs=mean)
        self.model = Model(inputs=[self.input_layer, self.sf_layer], outputs=mean)  # yuge
        self.arch_loss = tf.math.reduce_mean(tf.math.square(self.model.get_layer('z_fixed').get_weights()[0] - self.z_predicted))
        self.model.add_loss(self.arch_loss*self.lat_coef)

        self.encoder = self.get_encoder()
        
    def save(self):
        if self.file_path:
            os.makedirs(self.file_path, exist_ok=True)
            with open(os.path.join(self.file_path, 'model.pickle'), 'wb') as f:
                pickle.dump(self, f)

    def get_encoder(self, activation=False):
   
        return Model(inputs=self.model.input,
                    outputs=self.model.get_layer('center').output)
    
    def get_decoder(self):
        
        # Extract decoder fitted weights
        restored_w = []
        for i in range(self.dec_layer_num):
            restored_w.extend(self.model.get_layer('dec%s'%i).get_weights())
        restored_w.extend(self.model.get_layer('scaled_mean').get_weights())
        
        # Construct decoder
        dec_input_layer = Input(shape=(self.dim_latent_space,), name='latent_space')
        last_hidden = dec_input_layer
        for i, hid_size in enumerate(self.hidden_size[(self.center_idx+1):]):
            last_hidden = Dense(hid_size, activation=tf.nn.leaky_relu,
                                kernel_initializer=self.init,
                                kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                                name="dec%s"%i)(last_hidden)
            
        mean = Dense(self.output_size, kernel_initializer=self.init, activation="softmax",
                     kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                     name='scaled_mean')(last_hidden)
        
        dec_model = Model(inputs=dec_input_layer, outputs=mean)
        dec_model.set_weights(restored_w)
        
        return dec_model

    def predict(self, nor_count, lib_size, return_info = False):
        
        print('scAAnet: Calculating reconstructions...')
        preds = self.model.predict({'nor_count': nor_count,
                                    'lib_size': lib_size})
        if isinstance(preds, list):
            recon = preds[0]
        else:
            recon = preds

        print('scAAnet: Calculating low dimensional representations...')
        usage = self.encoder.predict({'nor_count': nor_count,
                                      'lib_size': lib_size})
        
        print('scAAnet: Calculating spectra in the original space')
        self.decoder = self.get_decoder()
        spectra = self.decoder.predict(self.model.get_layer('z_fixed').get_weights()[0])

        return {'recon': recon, 'usage': usage, 'spectra': spectra}
    
class PoissonAutoencoder(Autoencoder):

    def build_output(self):
        mean = Dense(self.output_size, activation="softmax", kernel_initializer=self.init,
                     kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                     name='scaled_mean')(self.decoder_output)
        output = ColwiseMultLayer([mean, self.sf_layer])
        self.loss = poisson_loss

        self.extra_models['mean_norm'] = Model(inputs=self.input_layer, outputs=mean)
        self.model = Model(inputs=[self.input_layer, self.sf_layer], outputs=output)
        self.arch_loss = tf.math.reduce_mean(tf.math.square(self.model.get_layer('z_fixed').get_weights()[0] - self.z_predicted))
        self.model.add_loss(self.arch_loss*self.lat_coef)

        self.encoder = self.get_encoder()
        
class ZIPoissonAutoencoder(Autoencoder):
    
    def build_dec(self):
        
        # pi
        last_hidden_pi = Dense(self.hidden_size[(self.center_idx+1)], 
                            kernel_initializer=self.init, 
                            activation=tf.nn.leaky_relu,
                            kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                            name='dec_pi0')(Concatenate()([self.encoder_output, self.sf_layer]))
        
        if len(self.hidden_size) > (self.center_idx+2):
            for i, hid_size in enumerate(self.hidden_size[(self.center_idx+2):]):
                last_hidden_pi = Dense(hid_size, activation=tf.nn.leaky_relu,
                                    kernel_initializer=self.init,
                                    kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                                    name="dec_pi%s"%(i+1))(last_hidden_pi)
        # Scaled mean
        last_hidden_mean = Dense(self.hidden_size[(self.center_idx+1)], 
                            kernel_initializer=self.init, 
                            activation=tf.nn.leaky_relu,
                            kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                            name='dec0')(self.encoder_output)
        
        if len(self.hidden_size) > (self.center_idx+2):
            for i, hid_size in enumerate(self.hidden_size[(self.center_idx+2):]):
                last_hidden_mean = Dense(hid_size, activation=tf.nn.leaky_relu,
                                    kernel_initializer=self.init,
                                    kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                                    name="dec%s"%(i+1))(last_hidden_mean)
        
        self.decoder_output_mean = last_hidden_mean
        self.decoder_output_pi = last_hidden_pi
        self.dec_layer_num = len(self.hidden_size) - self.center_idx - 1
        self.build_output()

    def build_output(self):

        pi = Dense(self.output_size, activation=None, kernel_initializer=self.init,
                       kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                       name='pi')(self.decoder_output_pi)
        mean = Dense(self.output_size, activation="softmax", kernel_initializer=self.init,
                     kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                     name='scaled_mean')(self.decoder_output_mean)
        output = ColwiseMultLayer([mean, self.sf_layer])
        
        self.loss = ZIPoisson_loss

        self.extra_models['mean_norm'] = Model(inputs=self.input_layer, outputs=mean)
        
        self.model = Model(inputs=[self.input_layer, self.sf_layer], 
                           outputs=[output, pi])
        self.arch_loss = tf.math.reduce_mean(tf.math.square(self.model.get_layer('z_fixed').get_weights()[0] - self.z_predicted))
        self.model.add_loss(self.arch_loss*self.lat_coef)

        self.encoder = self.get_encoder()

    def predict(self, nor_count, lib_size, return_info=False):
        
        preds = super().predict(nor_count, lib_size)

        if return_info:
            _, pi = self.model.predict({'nor_count': nor_count,'lib_size': lib_size})
            preds['pi'] = pi
        
        return preds
        
class NBAutoencoder(Autoencoder):
    
    def build_dec(self):
        
        # Dispersion
        last_hidden_disp = Dense(self.hidden_size[(self.center_idx+1)], 
                            kernel_initializer=self.init, 
                            activation=tf.nn.leaky_relu,
                            kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                            name='dec_disp0')(Concatenate()([self.encoder_output, self.sf_layer]))
        
        if len(self.hidden_size) > (self.center_idx+2):
            for i, hid_size in enumerate(self.hidden_size[(self.center_idx+2):]):
                last_hidden_disp = Dense(hid_size, activation=tf.nn.leaky_relu,
                                    kernel_initializer=self.init,
                                    kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                                    name="dec_disp%s"%(i+1))(last_hidden_disp)
        # Scaled mean
        last_hidden_mean = Dense(self.hidden_size[(self.center_idx+1)], 
                            kernel_initializer=self.init, 
                            activation=tf.nn.leaky_relu,
                            kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                            name='dec0')(self.encoder_output)
        
        if len(self.hidden_size) > (self.center_idx+2):
            for i, hid_size in enumerate(self.hidden_size[(self.center_idx+2):]):
                last_hidden_mean = Dense(hid_size, activation=tf.nn.leaky_relu,
                                    kernel_initializer=self.init,
                                    kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                                    name="dec%s"%(i+1))(last_hidden_mean)
        
        self.decoder_output_mean = last_hidden_mean
        self.decoder_output_disp = last_hidden_disp
        self.dec_layer_num = len(self.hidden_size) - self.center_idx - 1
        self.build_output()

    def build_output(self):
        
        if self.dispersion == 'gene-cell':
            disp = Dense(self.output_size, activation=DispAct,
                               kernel_initializer=self.init,
                               kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                               name='dispersion')(self.decoder_output_disp)
        else:
            disp = DispLayer(self.output_size)(self.decoder_output_disp)
            
        mean = Dense(self.output_size, activation="softmax", kernel_initializer=self.init,
                     kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                     name='scaled_mean')(self.decoder_output_mean)
        output = ColwiseMultLayer([mean, self.sf_layer])
        
        self.loss = NB_loss
        
        self.extra_models['mean_norm'] = Model(inputs=self.input_layer, outputs=mean)
        
        self.model = Model(inputs=[self.input_layer, self.sf_layer], 
                           outputs=[output, disp])
        self.arch_loss = tf.math.reduce_mean(tf.math.square(self.model.get_layer('z_fixed').get_weights()[0] - self.z_predicted))
        self.model.add_loss(self.arch_loss*self.lat_coef)

        self.encoder = self.get_encoder()

    def predict(self, nor_count, lib_size, return_info=False):
        
        preds = super().predict(nor_count, lib_size)

        if return_info:
            _, disp = self.model.predict({'nor_count': nor_count,'lib_size': lib_size},
                                          batch_size=nor_count.shape[0])
            preds['disp'] = disp
        
        return preds
        
class ZINBAutoencoder(Autoencoder):
    
    def build_dec(self):
        
        # Dispersion/pi
        last_hidden_disp_pi = Dense(self.hidden_size[(self.center_idx+1)], 
                            kernel_initializer=self.init, 
                            activation=tf.nn.leaky_relu,
                            kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                            name='dec_disp_pi0')(Concatenate()([self.encoder_output, self.sf_layer]))
        
        if len(self.hidden_size) > (self.center_idx+2):
            for i, hid_size in enumerate(self.hidden_size[(self.center_idx+2):]):
                last_hidden_disp_pi = Dense(hid_size, activation=tf.nn.leaky_relu,
                                    kernel_initializer=self.init,
                                    kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                                    name="dec_disp_pi%s"%(i+1))(last_hidden_disp_pi)
        # Scaled mean
        last_hidden_mean = Dense(self.hidden_size[(self.center_idx+1)], 
                            kernel_initializer=self.init, 
                            activation=tf.nn.leaky_relu,
                            kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                            name='dec0')(self.encoder_output)
        
        if len(self.hidden_size) > (self.center_idx+2):
            for i, hid_size in enumerate(self.hidden_size[(self.center_idx+2):]):
                last_hidden_mean = Dense(hid_size, activation=tf.nn.leaky_relu,
                                    kernel_initializer=self.init,
                                    kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                                    name="dec%s"%(i+1))(last_hidden_mean)
        
        self.decoder_output_mean = last_hidden_mean
        self.decoder_output_disp_pi = last_hidden_disp_pi
        self.dec_layer_num = len(self.hidden_size) - self.center_idx - 1
        self.build_output()

    def build_output(self):
        
        if self.dispersion == 'gene-cell':
            disp = Dense(self.output_size, activation=DispAct,
                               kernel_initializer=self.init,
                               kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                               name='dispersion')(self.decoder_output_disp_pi)
        else:
            disp = DispLayer(self.input_size)(self.decoder_output_disp_pi)
        
        pi = Dense(self.output_size, activation=None, kernel_initializer=self.init,
                       kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                       name='pi')(self.decoder_output_disp_pi)
        
        mean = Dense(self.output_size, activation='softmax', kernel_initializer=self.init,
                       kernel_regularizer=l1_l2(self.l1_coef, self.l2_coef),
                       name='scaled_mean')(self.decoder_output_mean)      
        output = ColwiseMultLayer([mean, self.sf_layer])

        self.loss = ZINB_loss

        self.extra_models['mean_norm'] = Model(inputs=self.input_layer, outputs=mean)

        self.model = Model(inputs=[self.input_layer, self.sf_layer], 
                           outputs=[output, disp, pi])
        self.arch_loss = tf.math.reduce_mean(tf.math.square(self.model.get_layer('z_fixed').get_weights()[0] - self.z_predicted))
        self.model.add_loss(self.arch_loss*self.lat_coef)

        self.encoder = self.get_encoder()

    def predict(self, nor_count, lib_size, return_info=False):

        # warning! this may overwrite adata.X
        preds = super().predict(nor_count, lib_size)
        
        if return_info:
            _, disp, pi = self.model.predict({'nor_count': nor_count, 'lib_size': lib_size}, 
                                       batch_size=nor_count.shape[0])
            preds['disp'] = disp
            preds['pi'] = pi
        
        return preds
        
AE_types = {'normal': Autoencoder, 'poisson': PoissonAutoencoder,
            'zipoisson': ZIPoissonAutoencoder,
            'nb': NBAutoencoder, 'zinb': ZINBAutoencoder}